from flatlib import const, aspects
from flatlib.chart import Chart
from flatlib.geopos import GeoPos
from flatlib.datetime import Datetime, Date
from flatlib.object import GenericObject

import collections
import polars as pl

from .utils import *


## GLOBAL VARS
ROMAN_HOUSE_NUMBERS = {'House1': 'I', 'House2': 'II', 'House3': 'III', 'House4': 'IV', 'House5': 'V', 'House6': 'VI',
                    'House7': 'VII', 'House8': 'VIII', 'House9': 'IX', 'House10': 'X', 'House11': 'XI', 'House12': 'XII'
                    }

RASHIS = ['Aries', 'Taurus', 'Gemini', 'Cancer', 'Leo', 'Virgo', 'Libra', 
          'Scorpio', 'Sagittarius', 'Capricorn', 'Aquarius', 'Pisces']

## Lords of the 12 Zodiac Signs
SIGN_LORDS = ["Mars", "Venus", "Mercury", "Moon", "Sun", "Mercury", "Venus", "Mars", "Jupiter", "Saturn", "Saturn", "Jupiter"]          

NAKSHATRAS = ['Ashwini','Bharani','Krittika','Rohini','Mrigashīrsha', 'Ardra', 'Punarvasu', 'Pushya', 'Āshleshā', 
'Maghā', 'PūrvaPhalgunī', 'UttaraPhalgunī', 'Hasta', 'Chitra', 'Svati', 'Vishakha', 'Anuradha', 'Jyeshtha', 'Mula', 
'PurvaAshadha','UttaraAshadha', 'Shravana', 'Dhanishta','Shatabhisha', 'PurvaBhādrapadā', 'UttaraBhādrapadā', 'Revati']

AYANAMSA_MAPPING = { "Lahiri": const.AY_LAHIRI, "Krishnamurti": const.AY_KRISHNAMURTI, "Raman": const.AY_RAMAN,
                    "FaganBradley": const.AY_FAGAN_BRADLEY, "Deluce": const.AY_DELUCE, "Sassanian": const.AY_SASSANIAN,
                    "Aldebaran15Taurus": const.AY_ALDEBARAN_15TAU, "GalacticCenter_05_Sag": const.AY_GALCENTER_5SAG
                    }


HOUSE_SYSTEM_MAPPING = { "Placidus": const.HOUSES_PLACIDUS, "Koch": const.HOUSES_EQUAL, 
                        "Porphyrius": const.HOUSES_PORPHYRIUS, "Regiomontanus": const.HOUSES_REGIOMONTANUS, 
                        "Campanus": const.HOUSES_CAMPANUS, "Equal": const.HOUSES_EQUAL,
                        "Equal 2": const.HOUSES_EQUAL_2, "Vehlow Equal": const.HOUSES_VEHLOW_EQUAL, 
                        "Whole Sign": const.HOUSES_WHOLE_SIGN, "Meridian": const.HOUSES_MERIDIAN, 
                        "Azimuthal": const.HOUSES_AZIMUTHAL, "Polich Page": const.HOUSES_POLICH_PAGE,
                        "Alcabitus": const.HOUSES_ALCABITUS, "Morinus": const.HOUSES_MORINUS
                    }

ASPECT_MAPPING = {  const.NO_ASPECT: "No Aspect", const.CONJUNCTION: "Conjunction", const.SEXTILE: "Sextile",
                    const.SQUARE: "Square", const.TRINE: "Trine", const.OPPOSITION: "Opposition", 
                    const.SEMISEXTILE: "Semi Sextile", const.SEMIQUINTILE: "Semi Quintile",
                    const.SEMISQUARE: "Semi Square", const.QUINTILE: "Quintile", 
                    const.SESQUIQUINTILE: "Sesqui Quintile", const.SESQUISQUARE: "Sesqui Square",
                    const.BIQUINTILE: "Bi Quintile", const.QUINCUNX: "Quincunx",
                    }

# Columns names for NamedTuple Collections / Final Output DataFrames
HOUSES_TABLE_COLS = ["Object", "HouseNr","Rasi", "LonDecDeg", "SignLonDMS", "SignLonDecDeg", "DegSize",
                     "Nakshatra", "RasiLord", "NakshatraLord", "SubLord", "SubSubLord"]

PLANETS_TABLE_COLS = ["Object", "Rasi", "isRetroGrade", "LonDecDeg", "SignLonDMS", "SignLonDecDeg", "LatDMS",
                        "Nakshatra", "RasiLord", "NakshatraLord", "SubLord", "SubSubLord" ,"HouseNr"]


class VedicHoroscopeData:
    def __init__(self, year:int, month:int, day:int, hour:int, minute:int, second : int, utc:str, 
                 latitude:float, longitude:float, ayanamsa: str = "Lahiri", house_system : str = "Equal"):
        """
        Generates Planetary and House Positions Data for a time and place input.

        Parameters
        ==========
        year: Birth Year
        month: Birth Month
        day: Birth Day
        house: Birth Hour
        minute: Birth Minute
        utc: Country UTC
        latitude: latitude
        longitude: longitude
        ayanamsa: ayanamsa 
        """
        self.year       = year
        self.month      = month
        self.day        = day
        self.hour       = hour
        self.minute     = minute
        self.second     = second
        self.utc        = utc
        self.latitude   = latitude
        self.longitude  = longitude
        self.ayanamsa   = ayanamsa
        self.house_system = house_system
    
    def get_ayanamsa(self):
        """Returns an Ayanamsa System from flatlib.sidereal library, based on user input"""
        return AYANAMSA_MAPPING.get(self.ayanamsa, None)
    
    def get_house_system(self):
        """Returns an House System from flatlib.sidereal library, based on user input"""
        return HOUSE_SYSTEM_MAPPING.get(self.house_system, None)

    def generate_chart(self):
        """Generates a `flatlib.Chart` object for the given time and location data"""
        date = Datetime([self.year, self.month, self.day], ["+",self.hour, self.minute, self.second], self.utc)
        geopos = GeoPos(self.latitude, self.longitude)
        chart = Chart(date, geopos, IDs=const.LIST_OBJECTS, hsys=self.get_house_system(), mode = self.get_ayanamsa())
        return chart

    def get_planetary_aspects(self, chart: Chart):
        """      """
        planets = [const.SUN, const.MOON, const.MARS, const.MERCURY, const.JUPITER, const.VENUS, const.SATURN,
                    const.URANUS, const.NEPTUNE, const.PLUTO, const.NORTH_NODE, const.SOUTH_NODE]
        aspects_output = []
        aspects_dict = []

        for p1 in planets:
            for p2 in planets:
                if p1 != p2:
                    obj1 = chart.get(p1)
                    obj2 = chart.get(p2)
                    aspect = aspects.getAspect(obj1, obj2, const.ALL_ASPECTS)
                    ## Replace North and South nodes with conventional names
                    p1_new = p1.replace("North Node", "Rahu").replace("South Node", "Ketu")
                    p2_new = p2.replace("North Node", "Rahu").replace("South Node", "Ketu")
                    if aspect.exists():
                        aspect_type = ASPECT_MAPPING[int(aspect.type)]  # Use global variable here
                        aspect_orb = round(aspect.orb, 3)  # get the orb value
                        aspects_output.append(f'{p1_new} and {p2_new} are in {aspect_type} aspect with an orb of {aspect_orb} degrees')
                        aspects_dict.append({"P1":p1_new, "P2": p2_new, "AspectType" : aspect_type, "AspectDeg" : aspect.type, "AspectOrb" : aspect_orb })

        return aspects_dict


    def get_ascendant_data(self, asc_data: GenericObject, PlanetsDataCollection : collections.namedtuple):
        """Generates Ascendant Data and returns the data in the format of the PlanetsDataCollection Named Tuple"""
        asc_chart_data = clean_select_objects_split_str(str(asc_data))
        asc_rl_nl_sl_data = self.get_rl_nl_sl_data(deg = asc_data.lon)
        # Create a dictionary with None values for all fields
        data_dict = {field: None for field in PlanetsDataCollection._fields}
        # Update the specific fields with the ascendant data
        data_dict["Object"] = asc_chart_data[0]
        data_dict["Rasi"] = asc_chart_data[1]
        data_dict["SignLonDMS"] = asc_chart_data[2]
        data_dict["Nakshatra"] = asc_rl_nl_sl_data.get("Nakshatra", None)
        data_dict["RasiLord"]  = asc_rl_nl_sl_data.get("RasiLord", None)
        data_dict["SubLord"] = asc_rl_nl_sl_data.get("SubLord", None)
        data_dict["SubSubLord"] = asc_rl_nl_sl_data.get("SubSubLord", None)
        data_dict["NakshatraLord"] = asc_rl_nl_sl_data.get("NakshatraLord", None)
        data_dict["isRetroGrade"] = None
        data_dict["LonDecDeg"] = round(asc_data.lon, 3)
        data_dict["SignLonDecDeg"] = dms_to_decdeg(asc_chart_data[2])
        data_dict["LatDMS"] = None
        data_dict["HouseNr"] = 1

        # Return a new PlanetsDataCollection instance with the data
        return PlanetsDataCollection(**data_dict)

    def get_rl_nl_sl_data(self, deg : float):
        """
        Returns the  Rashi (Sign) Lord, Nakshatra, Nakshatra Pada, Nakshatra Lord, Sub Lord and Sub Sub Lord 
        corresponding to the given degree.
        """
        duration = [7, 20, 6, 10, 7, 18, 16, 19, 17]

        lords = ["Ketu", "Venus", "Sun", "Moon", "Mars", "Rahu", "Jupiter", "Saturn", "Mercury"]

        star_lords = lords * 3 ## lords for the 27 Nakshatras

        ## Compute Sign lords
        sign_deg = deg % 360  # Normalize degree to [0, 360)
        sign_index = int(sign_deg // 30)  # Each zodiac sign is 30 degrees
        
        # Compute Nakshatra details
        nakshatra_deg = sign_deg % 13.332  # Each nakshatra is 13.332 degrees
        nakshatra_index = int(sign_deg // 13.332)  # Find the nakshatra index
        pada = int((nakshatra_deg % 13.332) // 3.325) + 1  # Each pada is 3.325 degrees

        # Ensure nakshatra_index is within bounds
        nakshatra_index = nakshatra_index % len(NAKSHATRAS)        

        # Compute SubLords
        deg = deg - 120 * int(deg / 120)
        degcum = 0
        i = 0

        while i < 9:
            deg_nl = 360 / 27
            j = i
            while True:
                deg_sl = deg_nl * duration[j] / 120
                k = j
                while True:
                    deg_ss = deg_sl * duration[k] / 120
                    degcum += deg_ss
                    if degcum >= deg:
                        return {"Nakshatra": NAKSHATRAS[nakshatra_index], "Pada": pada, 
                                "NakshatraLord": star_lords[nakshatra_index], "RasiLord": SIGN_LORDS[sign_index], 
                                "SubLord": lords[j], "SubSubLord": lords[k] }
                    k = (k + 1) % 9
                    if k == j:
                        break
                j = (j + 1) % 9
                if j == i:
                    break
            i += 1  
    

    def get_planets_data_from_chart(self, chart: Chart, new_houses_chart: Chart = None):
        """
        Generate the planets data table given a `flatlib.Chart` object.
        Parameters
        ==========
        chart: flatlib Chart object using which planetary positions have to be generated
        new_houses_chart: flatlib Chart Object using which new house numbers have to be
                        computed, typically used along with KP Horary Method
        """
        PlanetsData = collections.namedtuple("PlanetsData",PLANETS_TABLE_COLS)

        # Get the house each planet is in
        planet_in_house = self.get_planet_in_house(planets_chart = chart, houses_chart = new_houses_chart) if new_houses_chart \
                        else self.get_planet_in_house(planets_chart = chart, houses_chart = chart)

        ### Get Ascendant Data
        ascendant_data = self.get_ascendant_data(asc_data = chart.get(const.ASC), PlanetsDataCollection = PlanetsData)

        planets_data = []
        planets_data.append(ascendant_data)
        for planet in chart.objects:
            planet_obj = clean_select_objects_split_str(str(planet))
            planet_name, planet_lon_deg, planet_lat_deg = planet_obj[0], planet_obj[2], planet_obj[3]

            ## Get additional details like Nakshatra, RL, NL, SL details
            rl_nl_sl_data = self.get_rl_nl_sl_data(deg = planet.lon)
            planet_star = rl_nl_sl_data.get("Nakshatra", None)
            planet_rasi_lord = rl_nl_sl_data.get("RasiLord", None)
            planet_star_lord = rl_nl_sl_data.get("NakshatraLord", None)
            planet_sub_lord = rl_nl_sl_data.get("SubLord", None)
            planet_ss_lord = rl_nl_sl_data.get("SubSubLord", None)

            # Get the house the planet is in
            planet_house = planet_in_house.get(planet_name, None)

            ## Append data to NamedTuple Collection
            planets_data.append(PlanetsData(planet_name, planet.sign, planet.isRetrograde(), round(planet.lon,3), 
                                            planet_lon_deg, round(planet.signlon, 3), planet_lat_deg, planet_star, 
                                            planet_rasi_lord, planet_star_lord, planet_sub_lord, planet_ss_lord, planet_house))
        return planets_data

    def get_houses_data_from_chart(self, chart: Chart):
        """Generate the houses data table given a `flatlib.Chart` object"""
        HousesData = collections.namedtuple("HousesData", HOUSES_TABLE_COLS) # Create NamedTuple Collection to store data
        houses_data = []
        for house in chart.houses:    
            house_obj = str(house).strip('<').strip('>').split()
            house_name, house_lon_deg, house_size = house_obj[0], house_obj[2], round(float(house_obj[3]), 3)
            house_nr = int(house_name.strip("House"))
            house_roman_nr = ROMAN_HOUSE_NUMBERS.get(house_name)

            ## Get additional details like Nakshatra, RL, NL, SL details
            rl_nl_sl_data = self.get_rl_nl_sl_data(deg = house.lon)
            house_star = rl_nl_sl_data.get("Nakshatra", None)
            house_star_lord = rl_nl_sl_data.get("NakshatraLord", None)
            house_rasi_lord = rl_nl_sl_data.get("RasiLord", None)
            house_sub_lord = rl_nl_sl_data.get("SubLord", None)
            house_ss_lord = rl_nl_sl_data.get("SubSubLord", None)


            ## Append data to NamedTuple Collection
            houses_data.append(HousesData(house_roman_nr, house_nr, house.sign, round(house.lon,3), house_lon_deg, round(house.signlon, 3), 
                            house_size, house_star, house_rasi_lord, house_star_lord, house_sub_lord, house_ss_lord)) 
        return houses_data

    def get_consolidated_chart_data(self, planets_data: collections.namedtuple, houses_data: collections.namedtuple, 
                                    return_style : str = None):
        """
        Create consolidated dict data where all objects (both planets and houses) are listed by rasi (sign).
        If `return_style == "dataframe_records"`, returns the consolidated data in the form of a list of dictionaries
        If `return_style == None`, returns the consolidated data in the form of a dictionary grouped by rasi
        """
        # Construct polars DataFrame of planets and houses data from the flatlib_sidereal Chart object
        req_cols = ["Rasi","Object","isRetroGrade", "LonDecDeg" ,"SignLonDMS", "SignLonDecDeg"]
        planets_df = pl.DataFrame(planets_data).select(req_cols)
        houses_df = pl.DataFrame(houses_data).with_columns(pl.lit(False).alias("isRetroGrade")).select(req_cols)
        
        ## Create joined dataframe of planets and houses data
        df_concat = pl.concat([houses_df, planets_df])

        # Group by 'Rasi' and aggregate all columns into a list or first non-null value
        result_df = df_concat.group_by('Rasi').agg([
            pl.col('Object').map_elements(list, return_dtype=pl.Object).alias('Object'),
            pl.col('isRetroGrade').map_elements(list, return_dtype=pl.Object).alias('isRetroGrade'),
            pl.col('LonDecDeg').map_elements(list, return_dtype=pl.Object).alias('LonDecDeg'),
            pl.col('SignLonDMS').map_elements(list, return_dtype=pl.Object).alias('SignLonDMS'),
            pl.col('SignLonDecDeg').map_elements(list, return_dtype=pl.Object).alias('SignLonDecDeg'),
            # pl.col('LatDMS').map_elements(list, return_dtype=pl.Object).alias('LatDMS')
        ])

        ## Sort by Rashis Order from `Aries` to `Pisces` in Clockwise Order
        result_df = result_df.with_columns(pl.col('Rasi').map_elements(lambda rasi: RASHIS.index(rasi)).alias('RashiOrder'))
        result_df = result_df.sort('RashiOrder').drop('RashiOrder') ## Sort by RashiOrder and Drop the column


        if return_style == "dataframe_records":
            return result_df.to_dicts()
        else:
            return self.get_consolidated_chart_data_rasi_wise(df = result_df)

    def get_consolidated_chart_data_rasi_wise(self, df: pl.DataFrame):
        """Returns in dict format, the consolidated chart data stored in a polars DataFrame grouped by Rasi"""  
        final_dict = {}
        columns = df.columns

        for row in df.iter_rows():
            rasi = row[columns.index('Rasi')]
            final_dict[rasi] = {}           
            for obj, is_retrograde, lon_dd, lon_dms, sign_lon_dd in zip(row[columns.index('Object')], 
                                                                        row[columns.index('isRetroGrade')] , 
                                                                        row[columns.index('LonDecDeg')] , 
                                                                        row[columns.index('SignLonDMS')] , 
                                                                        row[columns.index('SignLonDecDeg')] ):

                final_dict[rasi][obj] = {"is_Retrograde": is_retrograde, "LonDecDeg": lon_dd, 
                                        "SignLonDMS" : lon_dms, "SignLonDecDeg": sign_lon_dd}
        return final_dict


    def get_planet_in_house(self, houses_chart: Chart, planets_chart: Chart):
        """Determine which house each planet is in given a `flatlib.Chart` object"""
        planet_in_house = {}

        # Get the list of cusps (the boundary between two houses) along with their house numbers
        cusps = sorted([(house.lon, int(house.id.replace('House', ''))) for house in houses_chart.houses])
        # Add the first cusp (plus 360 degrees) as the end of the last house
        cusps.append((cusps[0][0] + 360, cusps[0][1])) 
        # print("Cusps ADJ:",cusps)

        for planet in planets_chart.objects:
            planet_name = clean_select_objects_split_str(str(planet))[0]
            planet_lon = planet.lon
            for i in range(12):
                if cusps[i][0] <= planet_lon < cusps[i+1][0]:
                    planet_in_house[planet_name] = cusps[i][1]
                    break
                    ## Check if planet is overlapping into the last cusp
                elif cusps[i][0] <= planet_lon + 360 < cusps[i+1][0]:
                    planet_in_house[planet_name] = cusps[i][1]
                    break

        return planet_in_house
    
    def get_unique_house_nrs_for_rasi_lord(self, planets_df : pl.DataFrame, planet_name: str):
        """Returns the unique set of house numbers where the given planet is the rasi lord"""
        # Group by RasiLord and aggregate HouseNr into a list
        grouped_df = planets_df.group_by("RasiLord").agg([pl.col('HouseNr')\
                                .map_elements(list, return_dtype=pl.Object).alias('HouseNr')])

        # Filter the DataFrame for the given planet
        filtered_df = grouped_df.filter(pl.col('RasiLord') == planet_name)

        # If there are no matching rows, return an empty list
        if filtered_df.shape[0] == 0:
            return []

        # Get the list of house numbers for the given planet
        house_nrs = filtered_df['HouseNr'].to_list()[0]

        # Remove duplicates by converting the list to a set and then back to a list
        unique_house_nrs = list(set(house_nrs))

        return unique_house_nrs

    def get_planet_wise_significators(self, planets_data: collections.namedtuple, houses_data: collections.namedtuple):
        """Generate the ABCD significators table for each planet"""
        significators_table_cols = ["Planet", "A", "B", "C", "D"]
        SignificatorsData = collections.namedtuple("PlanetSignificators", significators_table_cols)

        # Get the planets data and Filter planets_data to remove objects like "Asc", "Chiron", "Syzygy", "Fortuna"
        planets_data = [planet for planet in planets_data if planet.Object not in ["Asc", "Chiron", "Syzygy", "Fortuna"]]

        # Get the house each planet is in
        planets_house_deposition = {data.Object: data.HouseNr for data in planets_data}


        significators_data = []
        for planet in planets_data:
            # A. House occupied by the star lord (Nakshatra Lord) of the planet
            A = planets_house_deposition.get(planet.NakshatraLord, None)

            # B. House occupied by the planet itself
            B = planet.HouseNr

            # C. House nrs where the star lord planet is also the rashi lord
            C = [data.HouseNr for data in houses_data if data.RasiLord == planet.NakshatraLord]

            # D. House nrs where the planet itself is also the rashi lord
            D = [data.HouseNr for data in houses_data if data.RasiLord == planet.Object]

            # Append data to NamedTuple Collection
            significators_data.append(SignificatorsData(planet.Object, A, B, C, D))

        return significators_data

    def get_house_wise_significators(self, planets_data : collections.namedtuple, houses_data: collections.namedtuple):
        """Generate the ABCD significators table for each house"""
        significators_table_cols = ["House", "A", "B", "C", "D"]
        SignificatorsData = collections.namedtuple("HouseSignificators", significators_table_cols)

        # Get the planets data and Filter planets_data to remove objects like "Asc", "Chiron", "Syzygy", "Fortuna"
        planets_data = [planet for planet in planets_data if planet.Object not in ["Asc", "Chiron", "Syzygy", "Fortuna"]]

        # Create a mapping of planets to their star lords (Nakshatra Lords)
        planet_to_star_lord = {data.Object: data.NakshatraLord for data in planets_data}

        significators_data = []
        for house in houses_data:
            # A. Planets in the star of occupants of that house
            # A1.1) Get all the rows which match with the house.HouseNr
            occupant_house_planets = [planet.Object for planet in planets_data if planet.HouseNr == house.HouseNr]
            # A1.2.) For each planet in occupant_house_planets find out if that planet is the Nakshatra Lord for any row in the planets_data
            A = [planet.Object for planet in planets_data if planet.NakshatraLord in occupant_house_planets]

            # B. Planets in that house
            B = [planet.Object for planet in planets_data if planet.HouseNr == house.HouseNr]

            # C. Planets in the star of owners of that house
            C = [planet for planet, star_lord in planet_to_star_lord.items() if star_lord == house.RasiLord]

            # D. Owner of that house
            D = house.RasiLord

            # Append data to NamedTuple Collection
            significators_data.append(SignificatorsData(house.Object, A, B, C, D))

        return significators_data
        

    def compute_vimshottari_dasa(self, chart: Chart):
        """Computes the Vimshottari Dasa for the chart"""
        # Get the moon object from the chart
        moon = chart.get(const.MOON)
        moon_details = clean_select_objects_split_str(str(moon))

        # Moon's Details
        moon_rl_nl_sl = self.get_rl_nl_sl_data(deg = moon.lon)
        moon_nakshatra = moon_rl_nl_sl["Nakshatra"]
        moon_nakshatra_lord = moon_rl_nl_sl["NakshatraLord"]
        moon_sign_lord = moon_rl_nl_sl["RasiLord"]
        # moon_sub_lord = moon_rl_nl_sl["SubLord"]

        # Helper function to format datetime objects to string
        dt_tuple_str = lambda start_date: start_date.strftime("%d-%m-%Y")

        # Define the sequence of the Dasa periods and their lengths
        dasa_sequence = ['Ketu', 'Venus', 'Sun', 'Moon', 'Mars', 'Rahu', 'Jupiter', 'Saturn', 'Mercury']
        dasa_lengths = [7, 20, 6, 10, 7, 18, 16, 19, 17]

        # Find the starting point of the Dasa sequence
        start_index = dasa_sequence.index(moon_nakshatra_lord)

        # Reorder the Dasa sequence to start with the moon's Nakshatra Lord
        dasa_sequence = dasa_sequence[start_index:] + dasa_sequence[:start_index]
        dasa_lengths = dasa_lengths[start_index:] + dasa_lengths[:start_index]
        dasa_order = dict(zip(dasa_sequence, dasa_lengths))

        # Compute the remaining portion of the starting Dasa
        typical_nakshatra_arc = 800 #degree - mins
        nakshatra_start = NAKSHATRAS.index(moon_nakshatra) * typical_nakshatra_arc
        moon_lon_mins = round(moon.lon * 60, 2) # dms_to_mins(moon_lon_dms)
        elapsed_moon_mins = moon_lon_mins - nakshatra_start
        remaining_arc_mins = typical_nakshatra_arc - elapsed_moon_mins
        starting_dasa_duration = dasa_order[moon_nakshatra_lord]
        start_dasa_remaining_duration = (starting_dasa_duration/typical_nakshatra_arc) * remaining_arc_mins
        start_dasa_elapsed_duration = starting_dasa_duration - start_dasa_remaining_duration

        # Compute the start and end times of each Maha Dasa and Bhukti
        vimshottari_dasa = {} # collections.OrderedDict()
        chart_date = (self.year, self.month, self.day, self.hour, self.minute)
        dasa_start_date = compute_new_date(start_date = chart_date, diff_value = start_dasa_elapsed_duration, direction = "backward")
        for i in range(len(dasa_sequence)):
            dasa = dasa_sequence[i]
            dasa_length = dasa_lengths[i]
            dasa_end_date = compute_new_date(start_date = tuple(dasa_start_date.timetuple())[:5], diff_value = dasa_length, direction = "forward")
            vimshottari_dasa[dasa] = {'start': dt_tuple_str(dasa_start_date), 'end': dt_tuple_str(dasa_end_date), 'bhuktis': {} } #collections.OrderedDict()
            bhukti_start_date = dasa_start_date

            # Find the starting point of the Dasa sequence
            start_index = dasa_sequence.index(dasa)            
            # Reorder the Bhukti sequence to start with the current main dasa            
            bhukti_sequence = dasa_sequence[start_index:] + dasa_sequence[:start_index]
            bhukti_lengths = dasa_lengths[start_index:] + dasa_lengths[:start_index]    
            for j in range(len(bhukti_sequence)):
                bhukti = bhukti_sequence[j]
                bhukti_length = dasa_length * bhukti_lengths[j] / 120  # The total length of all Bhuktis in a Maha Dasa is 120 years
                bhukti_end_date = compute_new_date(start_date = tuple(bhukti_start_date.timetuple())[:5], diff_value = bhukti_length, direction="forward")
                vimshottari_dasa[dasa]['bhuktis'][bhukti] = {'start':  dt_tuple_str(bhukti_start_date), 'end': dt_tuple_str(bhukti_end_date)}
                # vimshottari_dasa[dasa]['bhuktis'][bhukti] = collections.OrderedDict([('start',  dt_tuple_str(bhukti_start_date)), ('end', dt_tuple_str(bhukti_end_date))])
                bhukti_start_date = bhukti_end_date
            dasa_start_date = dasa_end_date

        return vimshottari_dasa


from datetime import datetime
from dateutil.relativedelta import relativedelta

## 
def clean_select_objects_split_str(input_str):
    """Rename and Clean certain chart objects like North, South Node and Fortuna"""
    cleaned_str = (input_str.strip('<').strip('>')
                          .replace("North Node", "Rahu")
                          .replace("South Node", "Ketu")
                          .replace("Pars Fortuna", "Fortuna"))
    return cleaned_str.split()

def pretty_data_table(named_tuple_data : list):
    """Converts a list of NamedTuple Collections to a PrettyTable"""
    from prettytable import PrettyTable
    # Create a PrettyTable instance
    table = PrettyTable()

    # Add field names (column headers)
    table.field_names = named_tuple_data[0]._fields 

    # Add rows
    for data in named_tuple_data:
        table.add_row(data)

    return table

def dms_to_decdeg(dms_str: str):
    """
    This function converts a string input in Degrees:Mins:Secs to Decimal Degrees
    """
    dms = dms_str.split(':')
    degrees = float(dms[0])
    minutes = float(dms[1])
    seconds = float(dms[2])
    return round(degrees + (minutes/60) + (seconds/3600), 4)


def dms_to_mins(dms_str: str):
    """
    This function converts a string input in Degrees:Mins:Secs to total minutes.
    """
    dms = dms_str.split(':')
    degrees = int(dms[0])
    minutes = int(dms[1])
    seconds = int(dms[2])
    total_minutes = degrees * 60 + minutes + seconds / 60
    return round(total_minutes, 2)  

def dms_difference(dms1_str: str, dms2_str: str):
    """
    This function computes the difference between two degree:mins:secs string objects
    and returns the difference as a degree:mins:secs string.
    """
    def dms_to_seconds(dms_str):
        dms = dms_str.split(':')
        degrees = int(dms[0])
        minutes = int(dms[1])
        seconds = int(dms[2])
        total_seconds = degrees * 3600 + minutes * 60 + seconds
        return total_seconds

    def seconds_to_dms(seconds):
        degrees = seconds // 3600
        seconds %= 3600
        minutes = seconds // 60
        seconds %= 60
        return f"{int(degrees)}:{int(minutes)}:{int(seconds)}"

    dms1_seconds = dms_to_seconds(dms1_str)
    dms2_seconds = dms_to_seconds(dms2_str)

    diff_seconds = abs(dms1_seconds - dms2_seconds)

    return seconds_to_dms(diff_seconds)


def convert_years_ymdhm(years):
    """
    This function converts decimal years into years, months, days, hours, and minutes.
    """
    # Constants
    months_per_year = 12
    days_per_month = 30  # Approximation
    hours_per_day = 24
    minutes_per_hour = 60

    # Compute the breakdown
    whole_years = int(years)
    months = (years - whole_years) * months_per_year
    whole_months = int(months)
    days = (months - whole_months) * days_per_month
    whole_days = int(days)
    hours = (days - whole_days) * hours_per_day
    whole_hours = int(hours)
    minutes = (hours - whole_hours) * minutes_per_hour
    whole_minutes = int(minutes)

    return whole_years, whole_months, whole_days, whole_hours, whole_minutes

def compute_new_date(start_date : tuple, diff_value : float, direction: str):
    """
    This function computes a new date and time given an initial date and time and a time difference.
    """
    # Unpack start_date and diff_params
    year, month, day, hour, minute = start_date
    years, months, days, hours, minutes = convert_years_ymdhm(diff_value)

    # Create initial datetime object
    initial_date = datetime(year, month, day, hour, minute)

    # Compute relativedelta object
    time_difference = relativedelta(years=years, months=months, days=days, hours=hours, minutes=minutes)

    # Compute new date
    if direction == 'backward':
        new_date = initial_date - time_difference
    elif direction == 'forward':
        new_date = initial_date + time_difference
    else:
        raise ValueError("direction must be either 'backward' or 'forward'")

    return new_date


import os
import math
import polars as pl
from typing import Tuple
from datetime import datetime
from flatlib import const
from flatlib.chart import Chart
from .VedicAstro import VedicHoroscopeData
from .utils import pretty_data_table, dms_to_decdeg

## Read KP SubLord Divisions CSV File
# Determine the absolute path to the directory where this script is located
current_dir = os.path.abspath(os.path.dirname(__file__))
csv_file_path = os.path.join(current_dir, "data", "KP_SL_Divisions.csv")
KP_SL_DMS_DATA = pl.read_csv(csv_file_path)
KP_SL_DMS_DATA = KP_SL_DMS_DATA\
                .with_columns(pl.arange(1, KP_SL_DMS_DATA.height + 1).alias("SL_Div_Nr"))\
                .with_columns([
                    pl.col('From_DMS').map_elements(dms_to_decdeg).alias('From_DecDeg'),
                    pl.col('To_DMS').map_elements(dms_to_decdeg).alias('To_DecDeg'),
                    pl.col("From_DMS").str.replace_all(":", "").cast(pl.Int32).alias("From_DMS_int"),
                    pl.col("To_DMS").str.replace_all(":", "").cast(pl.Int32).alias("To_DMS_int")
                ])

def calculate_asc_speed(chart_start: Chart, start_dt: datetime, utc: str, lat: float, lon: float, ayan: str, house_system: str):
    """Calculates the ascendant speed between two datetime objects."""
    
    # Generate asc data for the start moment
    asc_start = chart_start.get(const.ASC)
    asc_start_lon_deg = asc_start.lon
    
    # Generate asc data for the end moment
    vhd = VedicHoroscopeData(start_dt.year, start_dt.month, start_dt.day, start_dt.hour, start_dt.minute, start_dt.second + 59, 
                             utc, lat, lon, ayan, house_system)
    end_dt = datetime(vhd.year, vhd.month, vhd.day, vhd.hour, vhd.minute, vhd.second)
    chart_end = vhd.generate_chart()
    asc_end = chart_end.get(const.ASC)
    asc_end_lon_deg = asc_end.lon
    
    # Calculate the change in Ascendant position
    delta_degrees = (asc_end_lon_deg - asc_start_lon_deg) % 360
    # Calculate the time difference in seconds
    diff_seconds = (end_dt - start_dt).total_seconds()
    
    # Calculate the speed in degrees per second
    asc_speed = delta_degrees / diff_seconds
    # print(f"**Delta Degrees: {delta_degrees}°")
    # print(f"Time Delta for Asc Speed Calc: {diff_seconds} seconds")
    print(f"Asc Speed for {start_dt.date()} is {asc_speed}° per second")
    return round(asc_speed, 7)

def get_horary_ascendant_degree(horary_number: int):
    """
    Convert a horary number to ascendant degree of the starting subdivision
    """
    if 1 <= horary_number <= 249:
        row = KP_SL_DMS_DATA.filter(pl.col("SL_Div_Nr") == horary_number).select(["Sign", "From_DMS", "From_DecDeg"])
        data = row.to_dicts()[0]

        # Convert the sign to its starting degree in the zodiac circle
        sign_order = {'Aries': 0, 'Taurus': 30, 'Gemini': 60, 'Cancer': 90,
                    'Leo': 120, 'Virgo': 150, 'Libra': 180, 'Scorpio': 210,
                    'Sagittarius': 240, 'Capricorn': 270, 'Aquarius': 300, 'Pisces': 330 
                    }
        
        sign_start_degree = sign_order[data['Sign']]

        # Convert From_DecDeg to zodiac degree location
        zodiac_degree_location = sign_start_degree + data['From_DecDeg']
        data['ZodiacDegreeLocation'] = zodiac_degree_location
        return data
    else:
        return "SL Div Nr. out of range. Please provide a number between 1 and 249."



def find_exact_ascendant_time(year: int, month: int, day: int, hour: int, minute: int, second: int, utc: str, 
                              lat: float, lon: float, horary_ascendant: dict, house_system: str, ayan : str) -> Tuple:
    """
    Finds the exact time when the Ascendant is at the desired degree.

    Parameters:
    - year: year of the horary question (prasna)
    - month: month of the horary question
    - day: day of the horary question
    - hour: hour of the horary question
    - minute: minute of the horary question
    - second: second of the horary question
    - utc: The UTC offset of the horary question's location, i.e of the predictor (astrologer)
    - lat: Latitude pertaining to the horary question's predictor (astrologer)
    - lon: Longitude pertaining to the horary question's predictor (astrologer).
    - horary_ascendant: The desired Ascendant degree and sign to match.
    - house_system: The house system to use, as a byte string (e.g., b'P' for Placidus).
    - ayan: The ayanamsa to be used when constructing the chart


    Returns:
    Tuple of (matched_time), planets_data and houses_data, when the Ascendant matches the desired degree.
    If no match is found within the day, returns None.
    """
    horary_asc_lon_dec_deg = horary_ascendant["ZodiacDegreeLocation"]
    hora_horoscope = VedicHoroscopeData(year, month, day, hour, minute, second, utc, lat, lon, ayan, house_system)
    hora_chart = hora_horoscope.generate_chart()


    # Initial horoscope to get the midnight ascendant degree
    midnight_horoscope = VedicHoroscopeData(year, month, day, 0, 0, 0, utc, lat, lon, ayan, house_system)
    midnight_chart = midnight_horoscope.generate_chart()
    midnight_planets_data = midnight_horoscope.get_planets_data_from_chart(midnight_chart)
    midnight_asc_deg = midnight_planets_data[0].LonDecDeg

    # Calculate the degree difference correctly, considering the zodiac as a loop
    delta_degrees = (horary_asc_lon_dec_deg - midnight_asc_deg) % 360
    print(f"Midnight Asc Deg: {midnight_asc_deg} \nDegree Difference: {delta_degrees}")
    time_difference_in_seconds = delta_degrees / 0.00431 ## random val arrived by manually calculating the difference in asc pos over 2 datetimes for the same day

    ## Calculate Asc Speed for this particular day
    # start_time =  datetime(year, month, day, 0, 0, 0)
    # asc_speed = calculate_asc_speed(midnight_chart, start_time, utc, lat, lon, ayan, house_system)
    # time_difference_in_seconds = delta_degrees / asc_speed
    
    # Calculate approx hours and remaining seconds
    adj_hour, remaining_seconds = divmod(time_difference_in_seconds, 3600)

    # Calculate approx minutes and seconds (with seconds as a float for precision)
    adj_minute, adj_second = divmod(remaining_seconds, 60)

    # matched_time = f"{year}-{month}-{day} {int(adj_hour):02d}:{int(adj_minute):02d}:{adj_second:.3f}"
    # horoscope = VedicHoroscopeData(year, month, day, adj_hour, adj_minute, adj_second, utc, lat, lon, ayan, house_system)
    # final_chart = horoscope.generate_chart()
    # horary_planets_data = hora_horoscope.get_planets_data_from_chart(hora_chart, new_houses_chart = final_chart)
    # houses_data = horoscope.get_houses_data_from_chart(final_chart)
    # return matched_time, final_chart, horary_planets_data, houses_data

    # Start checking 5 minutes (300 seconds) before the calculated time
    start_hour = adj_hour
    start_minute = adj_minute        

    # Adjust for iterating over the rest of the day
    total_seconds = int((start_hour * 3600) + (start_minute * 60) ) # + adj_second 
    day_end_seconds = 24 * 3600  # Total seconds in a day


    for current_second in range(total_seconds, day_end_seconds):
        adjusted_hour = current_second // 3600
        remaining = current_second % 3600
        adjusted_minute = remaining // 60
        adjusted_second = remaining % 60
        horoscope = VedicHoroscopeData(year, month, day, adjusted_hour, adjusted_minute, adjusted_second, utc, lat, lon, ayan, house_system)
        chart_itr = horoscope.generate_chart()
        planets_data = horoscope.get_planets_data_from_chart(chart_itr)
        asc = planets_data[0]
        asc_lon_deg = asc.LonDecDeg

        if math.isclose(asc_lon_deg, horary_asc_lon_dec_deg, abs_tol=0.0018):
            horoscope = VedicHoroscopeData(year, month, day, adjusted_hour, adjusted_minute, adjusted_second+0.5, utc, lat, lon, ayan, house_system)
            final_chart = horoscope.generate_chart()       
            matched_time = f"{year:04d}-{month:02d}-{day:02d} {horoscope.hour}:{horoscope.minute}:{horoscope.second}"
            houses_data = horoscope.get_houses_data_from_chart(final_chart)
            horary_planets_data = hora_horoscope.get_planets_data_from_chart(hora_chart, new_houses_chart = final_chart)
            return matched_time, final_chart, horary_planets_data, houses_data

    print("No matching Ascendant time found for the given input")
    return None    

if __name__== "__main__":
    year = 2024
    month = 2
    day = 5
    hour = 9
    minute = 5
    secs = 0
    horary_number = 34
    latitude, longitude, utc = 11.020085773931049, 76.98319647719487, "+5:30" ## Coimbatore
    ayan = "Krishnamurti"
    house_system = "Placidus" ## Default House for KP system


    horary_asc = get_horary_ascendant_degree(horary_number)
    desired_asc = horary_asc["ZodiacDegreeLocation"]
    print(f'Ascendant Degrees: {desired_asc} degrees')
    matched_time, final_chart, planets_data, houses_data = find_exact_ascendant_time(year, month, day, hour, minute, secs, utc,
                                                                        latitude, longitude, horary_asc, house_system, ayan)

    if matched_time:
        print(f'The Ascendant is at {houses_data[0].LonDecDeg} degrees on {matched_time}')
        print("Planets Data:\n",pretty_data_table(planets_data))
        print("\nHouses Data:\n", pretty_data_table(houses_data))